# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xtrack as xt
import xpart as xp
import xwakes as xw
import xobjects as xo
from xobjects.test_helpers import for_all_test_contexts

import pytest
from scipy.constants import e as qe, c as c_light
import pandas as pd
import numpy as np
import pathlib
from xwakes.wit.component import KIND_DEFINITIONS as kind_to_parameters

test_data_folder = pathlib.Path(__file__).parent.joinpath(
    '../test_data').absolute()

exclude_contexts = ['ContextPyopencl', 'ContextCupy']


@for_all_test_contexts(excluding=exclude_contexts)
@pytest.mark.parametrize('wake_type', ['WakeResonator', 'WakeTable'],
                         ids=lambda x: x)
@pytest.mark.parametrize('kind', [
    ['longitudinal'],
    ['constant_x', 'constant_y'],
    ['dipolar_x','dipolar_y'],
    ['dipolar_xy','dipolar_yx'],
    ['quadrupolar_x','quadrupolar_y'],
    ['quadrupolar_xy','quadrupolar_yx'],
    ], ids=lambda x: '_'.join(x))
def test_wake_kick_single_bunch(test_context, kind, wake_type):
    # test single-bunch wake kick and check that the resonator wake definition
    # is consistent with the table wake definition.
    # in this test we place one particle in each slice, we offset some of them
    # and check that the wake kicks are consistent with the definitions
    p0c = 7000e9
    h_bunch = 600
    circumference=26658.883
    bucket_length = circumference/h_bunch/10
    zeta_range = (-0.5*bucket_length, 0.5*bucket_length)
    num_slices = 100

    particle_dummy = xp.Particles(p0c=p0c)

    if wake_type == 'WakeResonator':
        wf = xw.WakeResonator(
            kind=kind,
            r=1e8, q=1e5, f_r=1e3)
    else:
        wf_analytic = xw.WakeResonator(
                    kind=kind,
                    r=1e8, q=1e5, f_r=1e3)

        # would need beta_0 below here but we just need a larger value than the
        # revolution period
        t_table = np.linspace(0, circumference/c_light/particle_dummy.beta0[0], 1000)
        table = pd.DataFrame()
        table['time'] = t_table
        for i_comp, comp in enumerate(wf_analytic.components):
            print(kind[i_comp])
            table[kind[i_comp]] = comp.function_vs_t(t_table, particle_dummy.beta0, 1e-20)

        wf = xw.WakeFromTable(table)

    wf.configure_for_tracking(
        zeta_range=zeta_range,
        num_slices=num_slices,
        bunch_spacing_zeta=circumference/h_bunch,
        circumference=circumference
        )

    line = xt.Line(elements=[wf])
    line.build_tracker()
    line.particle_ref=xp.Particles(p0c=p0c)

    particles = xp.Particles(p0c=p0c,
                        zeta=wf.slicer.zeta_centers.flatten(),
                        weight=1e12,
                        _context=test_context)

    i_source_x = 5
    i_source_y = 10

    displace_x = 2e-3
    displace_y = 3e-3

    # we displace after a certain particle index so that we check also that
    # in the dipolar and quadrupolar cases there is no wake in front
    particles.x[i_source_x:] += displace_x
    particles.y[i_source_y:] += displace_y

    dict_p_bef = {}

    for kk in kind:
        if kk == 'longitudinal':
            dict_p_bef[kk] = ('delta', particles.delta.copy())
        elif kk.split('_')[1] == 'x' or kk.split('_')[1] == 'xy':
            dict_p_bef[kk] = ('px', particles.px.copy())
        elif kk.split('_')[1] == 'y' or kk.split('_')[1] == 'yx':
            dict_p_bef[kk] = ('py', particles.py.copy())
        else:
            raise ValueError('Invalid kind')

    line.track(particles)

    assert len(wf.components) == len(kind)

    for comp, kk in zip(wf.components, kind):
        if comp.plane == 'z':
            scale = -particles.q0**2 * qe**2 / (
                particles.p0c[0] * particles.beta0[0]* qe) * particles.weight[0]
        else:
            scale = particles.q0**2 * qe**2 / (
                particles.p0c[0] * particles.beta0[0]* qe) * particles.weight[0]
        assert comp.plane == kind_to_parameters[kk]['plane']
        assert comp.source_exponents == kind_to_parameters[kk]['source_exponents']
        assert comp.test_exponents == kind_to_parameters[kk]['test_exponents']

        expected = np.zeros_like(particles.zeta)

        for i_test, z_test in enumerate(particles.zeta):
            expected[i_test] += (particles.x[i_test]**comp.test_exponents[0] *
                                particles.y[i_test]**comp.test_exponents[1] *
                                np.dot(particles.x**comp.source_exponents[0] *
                                        particles.y**comp.source_exponents[1],
                                        comp.function_vs_zeta(z_test - particles.zeta,
                                                            beta0=particles.beta0[0],
                                                            dzeta=1e-12)) * scale)

        xo.assert_allclose(getattr(particles, dict_p_bef[kk][0]) - dict_p_bef[kk][1],
                           expected, rtol=1e-4, atol=1e-20)


@for_all_test_contexts(excluding=exclude_contexts)
@pytest.mark.parametrize('kind', [
    ['longitudinal'],
    ['constant_x', 'constant_y'],
    ['dipolar_x','dipolar_y'],
    ['dipolar_xy','dipolar_yx'],
    ['quadrupolar_x','quadrupolar_y'],
    ['quadrupolar_xy','quadrupolar_yx'],
    ], ids=lambda x: '_'.join(x))
def test_wake_kick_multi_bunch(test_context, kind):
    # test multi bunch wake kick.
    # in this test we place one particle in each slice (for two bunches),
    # we offset some of them and check that the wake kicks are consistent with 
    # the definitions
    p0c = 7000e9
    h_bunch = 600
    circumference=26658.883
    bucket_length = circumference/h_bunch/10
    zeta_range = (-0.5*bucket_length, 0.5*bucket_length)
    num_slices = 100

    filling_scheme = np.zeros(h_bunch)
    filling_scheme[0] = 1
    filling_scheme[1] = 1
    bunch_selection = [0, 1]

    wf = xw.WakeResonator(
        kind=kind,
        r=1e8, q=1e5, f_r=1e3)

    wf.configure_for_tracking(
        zeta_range=zeta_range,
        num_slices=num_slices,
        filling_scheme=filling_scheme,
        bunch_spacing_zeta=circumference/h_bunch,
        bunch_selection=bunch_selection,
        circumference=circumference
        )

    line = xt.Line(elements=[wf])
    line.build_tracker()
    line.particle_ref=xp.Particles(p0c=p0c)

    particles = xp.Particles(p0c=p0c,
                            zeta=wf.slicer.zeta_centers.flatten(),
                            weight=1e12,
                            _context=test_context)

    i_source_x = 5
    i_source_y = 10

    displace_x = 2e-3
    displace_y = 3e-3

    # we displace after a certain particle index so that we check also that
    # in the dipolar and quadrupolar cases there is no wake in front
    particles.x[i_source_x:] += displace_x
    particles.y[i_source_y:] += displace_y

    dict_p_bef = {}

    for kk in kind:
        if kk == 'longitudinal':
            dict_p_bef[kk] = ('delta', particles.delta.copy())
        elif kk.split('_')[1] == 'x' or kk.split('_')[1] == 'xy':
            dict_p_bef[kk] = ('px', particles.px.copy())
        elif kk.split('_')[1] == 'y' or kk.split('_')[1] == 'yx':
            dict_p_bef[kk] = ('py', particles.py.copy())
        else:
            raise ValueError('Invalid kind')

    line.track(particles)

    assert len(wf.components) == len(kind)

    for comp, kk in zip(wf.components, kind):
        if comp.plane == 'z':
            scale = -particles.q0**2 * qe**2 / (
                particles.p0c[0] * particles.beta0[0]* qe) * particles.weight[0]
        else:
            scale = particles.q0**2 * qe**2 / (
                particles.p0c[0] * particles.beta0[0]* qe) * particles.weight[0]
        assert comp.plane == kind_to_parameters[kk]['plane']
        assert comp.source_exponents == kind_to_parameters[kk]['source_exponents']
        assert comp.test_exponents == kind_to_parameters[kk]['test_exponents']

        expected = np.zeros_like(particles.zeta)

        for i_test, z_test in enumerate(particles.zeta):
            expected[i_test] += (particles.x[i_test]**comp.test_exponents[0] *
                                particles.y[i_test]**comp.test_exponents[1] *
                                np.dot(particles.x**comp.source_exponents[0] *
                                        particles.y**comp.source_exponents[1],
                                        comp.function_vs_zeta(z_test - particles.zeta,
                                                            beta0=particles.beta0[0],
                                                            dzeta=1e-12)) * scale)

        xo.assert_allclose(getattr(particles, dict_p_bef[kk][0]) - dict_p_bef[kk][1],
                           expected, rtol=1e-4, atol=1e-20)


@for_all_test_contexts(excluding=exclude_contexts)
def test_beam_loading_theorem(test_context):
    # this test is less complete than the previous ones but it tests explicitly
    # the beam loading theorem for the longitudinal wake on two particles
    p0c = 1.2e9
    h_RF = 600
    n_slices = 100
    circumference = 26658.883
    bucket_length = circumference/h_RF
    zeta_range = (-0.5*bucket_length, 0.5*bucket_length)
    dz = (zeta_range[1] - zeta_range[0])/n_slices
    zeta = np.linspace(zeta_range[0] + dz/2, zeta_range[1]-dz/2,
                       n_slices)

    i_source = -10
    i_test = 10

    particles = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        p0c=p0c,
        zeta=[zeta[i_test], zeta[i_source]],
        delta=[1e-3, 2e-3],
        weight=1e12,
        _context=test_context
    )

    delta_bef = particles.delta.copy()

    wf = xw.WakeResonator(kind='longitudinal', r=1e8, q=1e7, f_r=1e3)
    wf.configure_for_tracking(zeta_range=zeta_range, num_slices=n_slices)

    line = xt.Line(elements=[wf],
                   element_names=['wf'])
    line.build_tracker()
    line.track(particles, num_turns=1)

    scale = -particles.q0**2 * qe**2 / (qe * particles.p0c[0] * particles.beta0[0]
                                       ) * particles.weight[0]

    assert len(wf.components) == 1
    comp = wf.components[0]
    w_zero_plus = comp.function_vs_zeta(-1e-12, beta0=1, dzeta=1e-20)
    xo.assert_allclose(w_zero_plus, 62831.85307, atol=0, rtol=1e-6) # Regresssion test

    assert particles.zeta[1] > particles.zeta[0]
    xo.assert_allclose(particles.delta[1] - delta_bef[1], scale * w_zero_plus / 2, # beam loading theorem
                          rtol=1e-4, atol=0)

    # Resonator frequency chosen to have practically constant wake
    xo.assert_allclose(particles.delta[0] - delta_bef[0], scale * w_zero_plus * 3 / 2,  # 1 from particle in front
                        rtol=1e-4, atol=0)                               # 1/2 from particle itself


@for_all_test_contexts(excluding=exclude_contexts)
@pytest.mark.parametrize('kind', [
    ['longitudinal'],
    ['constant_x', 'constant_y'],
    ['dipolar_x','dipolar_y'],
    ['dipolar_xy','dipolar_yx'],
    ['quadrupolar_x','quadrupolar_y'],
    ['quadrupolar_xy','quadrupolar_yx'],
    ], ids=lambda x: '_'.join(x))
def test_wake_kick_multiturn(test_context, kind):
    # test multi-turn wake kick.
    # in this test we place one particle in each slice, we offset some of them
    # and check that the wake kicks are consistent with the definitions when
    # tracking with multi-turn wakes
    num_turns_wake = 2
    p0c = 1.2e9
    h_RF = 600
    n_slices = 100
    circumference = 26658.883
    bucket_length = circumference/h_RF
    zeta_range = (-0.5*bucket_length, 0.5*bucket_length)

    wf = xw.WakeResonator(kind=kind,
                          r=1e8, q=1e7, f_r=1e3)
    wf.configure_for_tracking(zeta_range=zeta_range, num_slices=n_slices,
                              num_turns=num_turns_wake, circumference=circumference)


    i_source = -10

    particles = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        p0c=p0c,
        zeta=wf.slicer.zeta_centers.flatten(),
        weight=1e12,
        _context=test_context
    )

    displace_x = 2e-3
    displace_y = 3e-3

    particles.x[i_source] += displace_x
    particles.y[i_source] += displace_y

    dict_p_bef = {}

    for kk in kind:
        if kk == 'longitudinal':
            dict_p_bef[kk] = ('delta', particles.delta.copy())
        elif kk.split('_')[1] == 'x' or kk.split('_')[1] == 'xy':
            dict_p_bef[kk] = ('px', particles.px.copy())
        elif kk.split('_')[1] == 'y' or kk.split('_')[1] == 'yx':
            dict_p_bef[kk] = ('py', particles.py.copy())
        else:
            raise ValueError('Invalid kind')

    line = xt.Line(elements=[wf],
                   element_names=['wf'])
    line.build_tracker()
    line.track(particles, num_turns=3)

    scale = particles.q0**2 * qe**2 / (
        particles.p0c[0] * particles.beta0[0]* qe) * particles.weight[0]

    wake_kwargs = {
        'beta0': particles.beta0[0],
        'dzeta': 1e-20,
    }

    for comp, kk in zip(wf.components, kind):
        if comp.plane == 'z':
            scale = -particles.q0**2 * qe**2 / (
                particles.p0c[0] * particles.beta0[0]* qe) * particles.weight[0]
        else:
            scale = particles.q0**2 * qe**2 / (
                particles.p0c[0] * particles.beta0[0]* qe) * particles.weight[0]
        assert comp.plane == kind_to_parameters[kk]['plane']
        assert comp.source_exponents == kind_to_parameters[kk]['source_exponents']
        assert comp.test_exponents == kind_to_parameters[kk]['test_exponents']

        expected = np.zeros_like(particles.zeta)

        for i_test, z_test in enumerate(particles.zeta):
            expected[i_test] += (
                particles.x[i_test]**comp.test_exponents[0] *
                particles.y[i_test]**comp.test_exponents[1] *
                np.dot(particles.x**comp.source_exponents[0] *
                particles.y**comp.source_exponents[1],
                comp.function_vs_zeta(z_test - particles.zeta,
                                      **wake_kwargs)) *
                scale * (num_turns_wake + 1))

            expected[i_test] += (
                particles.x[i_test]**comp.test_exponents[0] *
                particles.y[i_test]**comp.test_exponents[1] *
                np.dot(particles.x**comp.source_exponents[0] *
                       particles.y**comp.source_exponents[1],
                       comp.function_vs_zeta(z_test - particles.zeta -
                                             circumference, **wake_kwargs)) *
                scale * num_turns_wake)

        xo.assert_allclose(getattr(particles, dict_p_bef[kk][0]) - dict_p_bef[kk][1],
                           expected, rtol=1e-4, atol=1e-20)


@for_all_test_contexts(excluding=exclude_contexts)
@pytest.mark.parametrize('kind', [
    ['longitudinal'],
    ['constant_x', 'constant_y'],
    ['dipolar_x','dipolar_y'],
    ['dipolar_xy','dipolar_yx'],
    ['quadrupolar_x','quadrupolar_y'],
    ['quadrupolar_xy','quadrupolar_yx'],
    ], ids=lambda x: '_'.join(x))
def test_wake_kick_multiturn_reset(test_context, kind):
    # a more explicit test on multi-turn wake kick reset.
    # we track for n_turns_wake, set the weights to zero and track for
    # n_turns_wake-1 after which we track for another turn checking that no
    # kick is applied to the particles
    num_turns_wake = 4
    p0c = 1.2e9
    h_RF = 600
    n_slices = 100
    circumference = 26658.883
    bucket_length = circumference/h_RF
    zeta_range = (-0.5*bucket_length, 0.5*bucket_length)

    wf = xw.WakeResonator(kind=kind,
                          r=1e8, q=1e7, f_r=1e3)
    wf.configure_for_tracking(zeta_range=zeta_range, num_slices=n_slices,
                              num_turns=num_turns_wake, circumference=circumference)


    i_source = -10

    particles = xt.Particles(
        mass0=xt.PROTON_MASS_EV,
        p0c=p0c,
        zeta=wf.slicer.zeta_centers.flatten(),
        weight=1e12,
        _context=test_context
    )

    displace_x = 2e-3
    displace_y = 3e-3

    particles.x[i_source] += displace_x
    particles.y[i_source] += displace_y

    line = xt.Line(elements=[wf],
                   element_names=['wf'])
    line.build_tracker()
    line.track(particles, num_turns=num_turns_wake)

    particles.weight *= 0
    line.track(particles, num_turns=num_turns_wake-1)

    dict_p_bef = {}
    for kk in kind:
        if kk == 'longitudinal':
            dict_p_bef[kk] = ('delta', particles.delta.copy())
        elif kk.split('_')[1] == 'x' or kk.split('_')[1] == 'xy':
            dict_p_bef[kk] = ('px', particles.px.copy())
        elif kk.split('_')[1] == 'y' or kk.split('_')[1] == 'yx':
            dict_p_bef[kk] = ('py', particles.py.copy())
        else:
            raise ValueError('Invalid kind')

    line.track(particles, num_turns=1)

    for kk in kind:
        xo.assert_allclose(getattr(particles, dict_p_bef[kk][0]),
                           dict_p_bef[kk][1], rtol=1e-4, atol=1e-20)

@for_all_test_contexts(excluding=exclude_contexts)
@pytest.mark.parametrize('kind', [['longitudinal'],
                                  ['constant_x', 'constant_y'],
                                  ['dipolar_x','dipolar_y'],
                                  ['dipolar_xy','dipolar_yx'],
                                  ['quadrupolar_x','quadrupolar_y'],
                                  ['quadrupolar_xy','quadrupolar_yx']],
                                  ids=lambda x: '_'.join(x))
def test_wake_kick_multibunch_pipeline(test_context, kind):
    # test multi-bunch wake kick with pipeline
    # in this test we place one particle in each slice for a random number of
    # bunches on two communicating pipeline branches, we offset them
    # and check that the wake kicks are consistent with the definitions
    p0c = 1e9

    # Filling scheme
    n_slots = 100
    filling_scheme = np.array(np.floor(np.random.rand(n_slots)+0.1), dtype=int)
    while np.sum(filling_scheme) < 2:
        # if less than 2 slots are filled, try again
        filling_scheme = np.array(np.floor(np.random.rand(n_slots)+0.1), dtype=int)
    filling_scheme[0] = 1
    filled_slots = np.nonzero(filling_scheme)[0]
    n_bunches = len(filled_slots)
    n_bunches_0 = int(np.floor(n_bunches/2))
    bunch_selection_0 = np.arange(n_bunches_0, dtype=int)
    bunch_selection_1 = np.arange(n_bunches_0, n_bunches, dtype=int)

    print('initialising pipeline')
    comm = xt.pipeline.core.PipelineCommunicator()
    pipeline_manager = xt.PipelineManager(comm)
    pipeline_manager.add_particles(f'b0', 0)
    pipeline_manager.add_particles(f'b1', 0)
    pipeline_manager.add_element('wake')

    bunch_spacing = 25E-9*c_light
    sigma_zeta = bunch_spacing/20
    zeta_range = (-1.1*sigma_zeta, 1.1*sigma_zeta)
    num_slices = 1001
    dzeta = (zeta_range[1]-zeta_range[0])/num_slices
    zeta_slice_edges = np.linspace(zeta_range[0], zeta_range[1], num_slices+1)
    zeta_centers = zeta_slice_edges[:-1]+dzeta/2

    zeta_0 = []
    for bunch_number in bunch_selection_0:
        zeta_0.append(zeta_centers-filled_slots[bunch_number]*bunch_spacing)
    zeta_0 = np.hstack(zeta_0)

    print('Initialising particles')
    particles_0 = xt.Particles(p0c=p0c, zeta=zeta_0,
                                x=np.random.randn(len(zeta_0)),
                                y=np.random.randn(len(zeta_0)),
                                _context=test_context)
    particles_0.init_pipeline('b0')

    zeta_1 = []
    for bunch_number in bunch_selection_1:
        zeta_1.append(zeta_centers-filled_slots[bunch_number]*bunch_spacing)
    zeta_1 = np.hstack(zeta_1)
    particles_1 = xt.Particles(p0c=p0c, zeta=zeta_1,
                                x=np.random.randn(len(zeta_1)),
                                y=np.random.randn(len(zeta_1)),
                                _context=test_context)
    particles_1.init_pipeline('b1')

    i_source_x = 0
    i_source_y = 0

    displace_x_0 = 2e-3
    displace_y_0 = 3e-3
    displace_x_1 = 4e-3
    displace_y_1 = 5e-3

    # we displace after a certain particle index so that we check also that
    # in the dipolar and quadrupolar cases there is no wake in front
    particles_0.x[i_source_x:] += displace_x_0
    particles_0.y[i_source_y:] += displace_y_0

    particles_1.x[i_source_x:] += displace_x_1
    particles_1.y[i_source_y:] += displace_y_1

    print('Initialising wake')
    n_turns_wake = 1
    circumference = n_slots * bunch_spacing
    wake_table_name = (test_data_folder /
                    'headtail_format_table_hllhc/HLLHC_wake_flattop_nocrab.dat')

    wake_file_columns = ['time', 'longitudinal', 'dipolar_x', 'dipolar_y',
                            'quadrupolar_x', 'quadrupolar_y', 'dipolar_xy',
                            'quadrupolar_xy', 'dipolar_yx', 'quadrupolar_yx',
                            'constant_x', 'constant_y']
    components = kind
    wake_df = xw.read_headtail_file(wake_table_name,
                                    wake_file_columns)
    wf_0 = xw.WakeFromTable(wake_df, columns=components)
    wf_0.configure_for_tracking(zeta_range=zeta_range,
                                num_slices=num_slices,
                                bunch_spacing_zeta=bunch_spacing,
                                filling_scheme=filling_scheme,
                                bunch_selection=bunch_selection_0,
                                num_turns=n_turns_wake,
                                circumference=circumference
                                )
    wf_0._wake_tracker.init_pipeline(pipeline_manager=pipeline_manager,
                                        element_name='wake',
                                        partner_names=['b1'])

    wf_1 = xw.WakeFromTable(wake_df, columns=components)
    wf_1.configure_for_tracking(zeta_range=zeta_range,
                                num_slices=num_slices,
                                bunch_spacing_zeta=bunch_spacing,
                                filling_scheme=filling_scheme,
                                bunch_selection=bunch_selection_1,
                                num_turns=n_turns_wake,
                                circumference=circumference
                                )
    wf_1._wake_tracker.init_pipeline(pipeline_manager=pipeline_manager,
                                        element_name='wake',
                                        partner_names=['b0'])

    dict_p_bef_0 = {}
    dict_p_bef_1 = {}

    for kk in kind:
        if kk == 'longitudinal':
            dict_p_bef_0[kk] = ('delta', particles_0.delta.copy())
            dict_p_bef_1[kk] = ('delta', particles_1.delta.copy())
        elif kk.split('_')[1] == 'x' or kk.split('_')[1] == 'xy':
            dict_p_bef_0[kk] = ('px', particles_0.px.copy())
            dict_p_bef_1[kk] = ('px', particles_1.px.copy())
        elif kk.split('_')[1] == 'y' or kk.split('_')[1] == 'yx':
            dict_p_bef_0[kk] = ('py', particles_0.py.copy())
            dict_p_bef_1[kk] = ('py', particles_1.py.copy())
        else:
            raise ValueError('Invalid kind')

    print('Initialising lines')
    line_0 = xt.Line(elements=[wf_0])
    line_1 = xt.Line(elements=[wf_1])
    print('Initialising multitracker')
    line_0.build_tracker()
    line_1.build_tracker()
    multitracker = xt.PipelineMultiTracker(
        branches=[xt.PipelineBranch(line=line_0, particles=particles_0),
                  xt.PipelineBranch(line=line_1, particles=particles_1)])
    print('Tracking')
    pipeline_manager.verbose = True
    multitracker.track(num_turns=1)
    print('loading test data')
    particles_tot = xt.Particles.merge([particles_0, particles_1])

    for particles, dict_p_bef, wf in [(particles_0, dict_p_bef_0, wf_0),
                                      (particles_1, dict_p_bef_1, wf_1)]:

        assert len(wf.components) == len(kind)

        for comp, kk in zip(wf.components, kind):
            if comp.plane == 'z':
                scale = -particles.q0**2 * qe**2 / (
                    particles.p0c[0] * particles.beta0[0]* qe) * particles.weight[0]
            else:
                scale = particles.q0**2 * qe**2 / (
                    particles.p0c[0] * particles.beta0[0]* qe) * particles.weight[0]
            assert comp.plane == kind_to_parameters[kk]['plane']
            assert comp.source_exponents == kind_to_parameters[kk]['source_exponents']
            assert comp.test_exponents == kind_to_parameters[kk]['test_exponents']

            expected = np.zeros_like(particles.zeta)

            for i_test, z_test in enumerate(particles.zeta):
                expected[i_test] += (particles.x[i_test]**comp.test_exponents[0] *
                                        particles.y[i_test]**comp.test_exponents[1] *
                                        np.dot(particles_tot.x**comp.source_exponents[0] *
                                            particles_tot.y**comp.source_exponents[1],
                                            comp.function_vs_zeta(z_test - particles_tot.zeta,
                                                                    beta0=particles_tot.beta0[0],
                                                                    dzeta=1e-12)) * scale)

            xo.assert_allclose(getattr(particles, dict_p_bef[kk][0]) - dict_p_bef[kk][1],
                    expected, rtol=1e-4, atol=1e-20)




# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import pandas as pd
from .basewake import BaseWake
from .wit import ComponentFromArrays
from .wit.component import KIND_DEFINITIONS


class WakeFromTable(BaseWake):

    def __init__(self, table, columns=None):

        if isinstance(table, dict):
            table = pd.DataFrame(table)

        self.table = table

        assert 'time' in table.columns

        columns = columns or table.columns
        columns = list(columns)
        if 'time' in columns:
            columns.remove('time')

        components = []
        for cc in columns:
            assert cc in table.columns, f'Column {cc} not in table'
            assert cc in KIND_DEFINITIONS, f'Invalid component {cc}'

            component = ComponentFromArrays(
                interpolation_times=table['time'].values,
                wake_samples=table[cc].values,
                kind=cc)
            components.append(component)

        self.components = components
        self.columns = columns




# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import numpy as np

from xfields.beam_elements.element_with_slicer import ElementWithSlicer
import json
import os
import xfields as xf
import xpart as xp

COORDS = ['x', 'px', 'y', 'py', 'zeta', 'delta']
SECOND_MOMENTS = {}
for c1 in COORDS:
    for c2 in COORDS:
        if c1 + '_' + c2 in SECOND_MOMENTS or c2 + '_' + c1 in SECOND_MOMENTS:
            continue
        SECOND_MOMENTS[c1 + '_' + c2] = (c1, c2)


class CollectiveMonitor(ElementWithSlicer):
    """
    A class to monitor the collective motion of the beam. The monitor can save
    bunch-by-bunch, slice-by-slice and particle-by-particle data. For the
    particle-by-particle data, a mask can be used to select the particles to
    be monitored. The monitor collects a buffer of `flush_data_every` turns
    before saving the data to disk, in order to reduce the number of I/O and
    avoid storing too much data in memory.

    Parameters
    ----------
    monitor_bunches : Bool
        If True, the bunch-by-bunch data is monitored
    monitor_slices : Bool
        If True, the slice-by-slice data is monitored
    monitor_particles : Bool
        If True, the particle-by-particle data is monitored
    base_file_name : str
        Base file name for the output files. If it is not specified, the data
        will not be saved to disk.
    particle_monitor_mask : np.ndarray
        Mask identifying the particles to be monitored. If later on we try to
        monitor a number of particles different than the length of the mask, an
        error will be raised.
    flush_data_every: int
        Number of turns after which the data is saved to disk
    stats_to_store : list
        List of the statistics to store for the bunch-by-bunch and
        slice-by-slice data
    stats_to_store_particles : list
        List of the statistics to store for the particle-by-particle data
    backend: str
        Backend used to save the data. For the moment only 'hdf5' and 'json'
        are supported, with 'hdf5' being the default
    zeta_range : Tuple
        Zeta range for each bunch used in the underlying slicer.
    num_slices : int
        Number of slices per bunch used in the underlying slicer. It should be
        specified if the slice-by-slice data is monitored.
    bunch_spacing_zeta : float
        Bunch spacing in meters.
    filling_scheme: np.ndarray
        List of zeros and ones representing the filling scheme. The length
        of the array is equal to the number of slots in the machine and each
        element of the array holds a one if the slot is filled or a zero
        otherwise.
    bunch_selection: np.ndarray
        List of the bunches indicating which slots from the filling scheme are
        used (not all the bunches are used when using multi-processing)
    _flatten: bool
        Use flattened wakes
    """

    _stats_to_store = [
        'mean_x', 'mean_px', 'mean_y', 'mean_py', 'mean_zeta',
        'mean_delta', 'sigma_x', 'sigma_y', 'sigma_zeta',
        'sigma_px', 'sigma_py', 'sigma_delta',
        'epsn_x', 'epsn_y', 'epsn_zeta', 'num_particles']

    _stats_to_store_particles = [
        'x', 'px', 'y', 'py', 'zeta', 'delta', 'weight', 'state']

    # Mapping from the stats to store in the monitor to the moments to store in
    # the slicer
    stat_to_slicer_moments = {
        'mean_x': ['x'],
        'mean_px': ['px'],
        'mean_y': ['y'],
        'mean_py': ['py'],
        'mean_zeta': ['zeta'],
        'mean_delta': ['delta'],
        'sigma_x': ['x', 'x_x'],
        'sigma_y': ['y', 'y_y'],
        'sigma_zeta': ['zeta', 'zeta_zeta'],
        'sigma_px': ['px', 'px_px'],
        'sigma_py': ['py', 'py_py'],
        'sigma_delta': ['delta', 'delta_delta'],
        'epsn_x': ['x', 'px', 'x_x', 'px_px', 'x_px'],
        'epsn_y': ['y', 'py', 'y_y', 'py_py', 'y_py'],
        'epsn_zeta': ['zeta', 'delta', 'delta_delta', 'zeta_zeta',
                      'zeta_delta'],
        'num_particles': ['num_particles']
    }

    def __init__(self,
                 monitor_bunches,
                 monitor_slices,
                 monitor_particles,
                 base_file_name=None,
                 particle_monitor_mask=None,
                 flush_data_every=1,
                 zeta_range=None,
                 num_slices=1,
                 bunch_spacing_zeta=None,
                 filling_scheme=None,
                 bunch_selection=None,
                 stats_to_store=None,
                 stats_to_store_particles=None,
                 backend='hdf5',
                 _flatten=False):

        slicer_moments = []
        if stats_to_store is not None:
            for stat in stats_to_store:
                assert stat in self._stats_to_store
            self.stats_to_store = stats_to_store
        else:
            self.stats_to_store = self._stats_to_store

        if stats_to_store_particles is not None:
            for stat in stats_to_store_particles:
                assert stat in self._stats_to_store_particles
            self.stats_to_store_particles = stats_to_store_particles
        else:
            self.stats_to_store_particles = self._stats_to_store_particles

        for stat in self.stats_to_store:
            # For each stat we store, we add the corresponding moment
            if stat == 'num_particles':
                slicer_moments.append(stat)
            else:
                slicer_moments.extend(self.stat_to_slicer_moments[stat])

        if backend == 'hdf5':
            self.extension = 'h5'
            self.flush_buffer_to_file_func = flush_buffer_to_file_hdf5
        elif backend == 'json':
            self.extension = 'json'
            self.flush_buffer_to_file_func = flush_buffer_to_file_json
        else:
            raise ValueError('Only hdf5 and json backends are supported.')

        self.base_file_name = base_file_name
        self.bunch_buffer = None
        self.slice_buffer = None
        self.particle_buffer = None
        self.monitor_bunches = monitor_bunches
        self.monitor_slices = monitor_slices
        self.monitor_particles = monitor_particles
        self.flush_data_every = flush_data_every
        self.particle_monitor_mask = particle_monitor_mask

        if base_file_name is not None:
            self._bunches_filename = (self.base_file_name + '_bunches.' +
                                    self.extension)

            if os.path.exists(self._bunches_filename) and monitor_bunches:
                os.remove(self._bunches_filename)

            self._slices_filename = (self.base_file_name + '_slices.' +
                                    self.extension)

            if os.path.exists(self._slices_filename) and monitor_slices:
                os.remove(self._slices_filename)

            self._particles_filename = (self.base_file_name + '_particles.' +
                                        self.extension)

            if os.path.exists(self._particles_filename) and monitor_particles:
                os.remove(self._particles_filename)

        self.pipeline_manager = None

        self.i_turn = 0

        super().__init__(
            slicer_moments=slicer_moments,
            zeta_range=zeta_range,  # These are [a, b] in the paper
            num_slices=num_slices,  # Per bunch, this is N_1 in the paper
            bunch_spacing_zeta=bunch_spacing_zeta,  # This is P in the paper
            filling_scheme=filling_scheme,
            bunch_selection=bunch_selection,
            with_compressed_profile=False
        )

    def _reconfigure_for_parallel(self, n_procs, my_rank):
        filled_slots = self.slicer.filled_slots
        scheme = np.zeros(np.max(filled_slots) + 1,
                          dtype=np.int64)
        scheme[filled_slots] = 1

        split_scheme = xp.matched_gaussian.split_scheme
        bunch_selection_rank = split_scheme(filling_scheme=scheme,
                                            n_chunk=int(n_procs))

        self.slicer = xf.UniformBinSlicer(
            filling_scheme=scheme,
            bunch_selection=bunch_selection_rank[my_rank],
            zeta_range=self.slicer.zeta_range,
            num_slices=self.slicer.num_slices,
            bunch_spacing_zeta=self.slicer.bunch_spacing_zeta,
            moments=self.slicer.moments
        )

    def track(self, particles, _slice_result=None, _other_bunch_slicers=None):
        super().track(particles=particles,
                      _slice_result=_slice_result,
                      _other_bunch_slicers=_other_bunch_slicers
                      )

        self.i_turn += 1

        if self.monitor_bunches:
            if self.bunch_buffer is None:
                self.bunch_buffer = self._init_bunch_buffer()

            self._update_bunch_buffer(particles)

            if (self.i_turn % self.flush_data_every == 0 and
                    self.base_file_name is not None):
                self.flush_buffer_to_file_func(
                    self.bunch_buffer, self._bunches_filename)
                self.bunch_buffer = self._init_bunch_buffer()

        if self.monitor_slices:
            if self.slice_buffer is None:
                self.slice_buffer = self._init_slice_buffer()

            self._update_slice_buffer(particles)

            if (self.i_turn % self.flush_data_every == 0 and
                    self.base_file_name is not None):
                self.flush_buffer_to_file_func(
                    self.slice_buffer, self._slices_filename)
                self.slice_buffer = self._init_slice_buffer()

        if self.monitor_particles:
            if self.particle_buffer is None:
                self.particle_buffer = self._init_particle_buffer()

            self._update_particle_buffer(particles)

            if (self.i_turn % self.flush_data_every == 0 and
                    self.base_file_name is not None):
                self.flush_buffer_to_file_func(
                    self.particle_buffer, self._particles_filename)
                self.particle_buffer = self._init_particle_buffer()

    def _init_bunch_buffer(self):
        buf = {}
        for bid in self.slicer.bunch_selection:
            # Convert to int to avoid json serialization issues
            buf[int(bid)] = {}
            for stats in self.stats_to_store:
                buf[int(bid)][stats] = np.zeros(self.flush_data_every)

        return buf

    def _init_slice_buffer(self):
        buf = {}
        for bid in self.slicer.bunch_selection:
            # Convert to int to avoid json serialization issues
            buf[int(bid)] = {}
            for stats in self.stats_to_store:
                buf[int(bid)][stats] = np.zeros((self.flush_data_every,
                                                 self.slicer.num_slices))

        return buf

    def _init_particle_buffer(self):
        num_particles = np.sum(self.particle_monitor_mask)
        buf = {}

        for stat in self.stats_to_store_particles:
            buf[stat] = np.zeros((self.flush_data_every, num_particles))

        return buf

    def _update_bunch_buffer(self, particles):
        i_bunch_particles = self._slice_result['i_slot_particles']

        for i_bunch, bid in enumerate(self.slicer.bunch_selection):
            bunch_mask = (i_bunch_particles == bid)
            beta = np.mean(particles.beta0[bunch_mask])
            gamma = 1 / np.sqrt(1 - beta**2)
            beta_gamma = beta * gamma
            for stat in self.stats_to_store:
                if stat == 'num_particles':
                    if len(self.slicer.bunch_selection) > 1:
                        val = np.sum(self.slicer.num_particles[i_bunch, :])
                    else:
                        val = np.sum(self.slicer.num_particles)
                else:
                    mom = getattr(particles, stat.split('_')[-1])
                    if stat.startswith('mean'):
                        val = np.mean(mom[bunch_mask])
                    elif stat.startswith('sigma'):
                        val = np.std(mom[bunch_mask])
                    elif stat.startswith('epsn'):
                        if stat.split('_')[-1] != 'zeta':
                            mom_p_str = 'p' + stat.split('_')[-1]
                        else:
                            mom_p_str = 'delta'
                        mom_p = getattr(particles, mom_p_str)
                        val = (np.sqrt(np.linalg.det(np.cov(mom[bunch_mask],
                                                            mom_p[bunch_mask])))
                               * beta_gamma)
                    elif stat == 'num_particles':
                        val = np.sum(self.slicer.num_particles[i_bunch, :])
                    else:
                        raise ValueError('Unknown statistics f{stat}')

                # here we need to use self.i_turn - 1 because the turn is
                # incremented before calling this method
                self.bunch_buffer[bid][stat][(self.i_turn - 1) %
                                             self.flush_data_every] = val

    def _update_slice_buffer(self, particles):
        i_bunch_particles = self._slice_result['i_slot_particles']

        for i_bunch, bid in enumerate(self.slicer.bunch_selection):
            bunch_mask = (i_bunch_particles == bid)
            # we use the bunch beta_gamma to calculate the emittance, while
            # we should use the slice beta_gamma
            beta = np.mean(particles.beta0[bunch_mask])
            gamma = 1 / np.sqrt(1 - beta**2)
            beta_gamma = beta * gamma
            for stat in self.stats_to_store:
                mom_str = stat.split('_')[-1]
                if stat.startswith('mean'):
                    if len(self.slicer.bunch_selection) > 1:
                        val = self.slicer.mean(mom_str)[i_bunch, :]
                    else:
                        val = self.slicer.mean(mom_str)
                elif stat.startswith('sigma'):
                    if len(self.slicer.bunch_selection) > 1:
                        val = self.slicer.std(mom_str)[i_bunch, :]
                    else:
                        val = self.slicer.std(mom_str)
                elif stat.startswith('epsn'):
                    if mom_str != 'zeta':
                        mom_p_str = 'p' + stat.split('_')[-1]
                    else:
                        mom_p_str = 'delta'

                    if len(self.slicer.bunch_selection) > 1:
                        val = (np.sqrt(self.slicer.var(mom_str)[i_bunch, :] *
                                    self.slicer.var(mom_p_str)[i_bunch, :] -
                                    self.slicer.cov(mom_str, mom_p_str)[i_bunch,
                                    :]) *
                            beta_gamma)
                    else:
                        val = (np.sqrt(self.slicer.var(mom_str) *
                                    self.slicer.var(mom_p_str) -
                                    self.slicer.cov(mom_str, mom_p_str)) *
                            beta_gamma)
                elif stat == 'num_particles':
                    if len(self.slicer.bunch_selection) > 1:
                        val = self.slicer.num_particles[i_bunch, :]
                    else:
                        val = self.slicer.num_particles
                else:
                    raise ValueError('Unknown statistics f{stat}')

                # here we need to use self.i_turn - 1 because the turn is
                # incremented before calling this method
                self.slice_buffer[bid][stat][(self.i_turn - 1) %
                                             self.flush_data_every, :] = val

    def _update_particle_buffer(self, particles):
        for stat in self.stats_to_store_particles:
            if (self.particle_monitor_mask is not None
                and len(self.particle_monitor_mask) != len(particles.x)):
                raise ValueError('The length of the particle monitor mask is '
                                 'different from the number of particles being tracked')
            val = getattr(particles, stat)[self.particle_monitor_mask]
            # here we need to use self.i_turn - 1 because the turn is
            # incremented before calling this method
            self.particle_buffer[stat][(self.i_turn - 1) %
                                       self.flush_data_every, :] = val


def flush_buffer_to_file_hdf5(buffer, filename):
    try:
        import h5py
    except ImportError:
        raise ImportError('h5py is required to save the data in hdf5 format')
    # Iif the file does not exist, create it and write the buffer
    if not os.path.exists(filename):
        with h5py.File(filename, 'w') as f:
            for key in buffer:
                if type(buffer[key]) is not dict:
                    if len(buffer[key].shape) == 1:
                        f.create_dataset(str(key),
                                         data=buffer[key],
                                         chunks=True, maxshape=(None,))
                    elif len(buffer[key].shape) == 2:
                        f.create_dataset(str(key),
                                         data=buffer[key],
                                         chunks=True,
                                         maxshape=(None, buffer[key].shape[1]))
                    else:
                        raise ValueError('The shape of the buffer is not '
                                         'supported')
                else:
                    # this is for the bunch or slice buffer
                    for subkey in buffer[key]:
                        if len(buffer[key][subkey].shape) == 1:
                            f.create_dataset(str(key) + '/' + str(subkey),
                                             data=buffer[key][subkey],
                                             chunks=True, maxshape=(None,))
                        elif len(buffer[key][subkey].shape) == 2:
                            f.create_dataset(str(key) + '/' + str(subkey),
                                             data=buffer[key][subkey],
                                             chunks=True,
                                             maxshape=(None,
                                                       buffer[key][subkey].shape[1]))
                        else:
                            raise ValueError('The shape of the buffer is not '
                                             'supported')

    # If the file exists, append the buffer
    else:
        with h5py.File(filename, 'a') as f:
            for key in buffer:
                if type(buffer[key]) is not dict:
                    # this is for the particle buffer
                    f[str(key)].resize((f[str(key)].shape[0] +
                                        buffer[key].shape[0]), axis=0)
                    f[str(key)][-buffer[key].shape[0]:] = buffer[key]
                else:
                    # this is for the bunch or slice buffer
                    for subkey in buffer[key]:
                        f[str(key) + '/' + str(subkey)].resize(
                            (f[str(key) + '/' + str(subkey)].shape[0] +
                             buffer[key][subkey].shape[0]), axis=0)
                        val = buffer[key][subkey]
                        f[str(key) + '/' +
                          str(subkey)][-buffer[key][subkey].shape[0]:] = val


def flush_buffer_to_file_json(buffer, filename):
    # To prepare the buffer for json, we need to convert the numpy arrays to
    # lists
    buffer_lists = {}
    for key in buffer:
        if type(buffer[key]) is not dict:
            # this is for the particle buffer
            buffer_lists[str(key)] = buffer[key].tolist()
        else:
            # this is for the bunch or slice buffer
            buffer_lists[str(key)] = {}
            for subkey in buffer[key]:
                buffer_lists[str(key)][subkey] = buffer[key][subkey].tolist()

    # Now we can write the buffer to the file
    if not os.path.exists(filename):
        with open(filename, 'w') as f:
            json.dump(buffer_lists, f, indent=2)
    else:
        with open(filename, 'r') as f:
            old_buffer = json.load(f)
        for key in buffer_lists:
            if type(buffer_lists[key]) is not dict:
                # this is for the particle buffer
                old_buffer[key] = np.concatenate((old_buffer[key], buffer_lists[key]), axis=0).tolist()
            else:
                # this is for the bunch or slice buffer
                for subkey in buffer_lists[key]:
                    old_buffer[key][subkey] = np.concatenate(
                        (old_buffer[key][subkey], buffer_lists[key][subkey]),
                        axis=0).tolist()

        with open(filename, 'w') as f:
            json.dump(old_buffer, f, indent=2)



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import numpy as np

import xpart as xp
import xfields as xf
import xtrack as xt
from xfields.slicers.compressed_profile import CompressedProfile


class TransverseDamper(xt.BeamElement):
    """
    A simple bunch-by-bunch transverse Damper implementation.

    Parameters
    ----------
    gain_x : float
        the horizontal damper gain in 1/turns (corresponding to a damping
        rate of gain_x/2).
    gain_y : float
        the vertical damper gain in 1/turns (corresponding to a damping rate
        of gain_x/2).
    zeta_range : tuple
        the range of zetas covered by the underlying slicer.
    num_slices : int
        the number of slices per bunch used by the underlying slicer.
    filling_scheme : np.ndarray, optional
        an array of zeros and ones representing the filling scheme. Only
        needed for multi-bunch tracking.
    bunch_selection : np.ndarray, optional
        an array indicating which slot each bunch occupies in the filling
        scheme. Only needed for multi-bunch tracking
    bunch_spacing_zeta : float, optional
        the bunch spacing in meters. Only needed for multi-bunch tracking
    circumference : float, optional
        the machine circumference. Only needed for multi-bunch tracking.

    """

    iscollective = True

    def __init__(self, gain_x, gain_y, zeta_range, num_slices,
                 circumference=None, bunch_spacing_zeta=None, filling_scheme=None,
                 bunch_selection=None, **kwargs):
        self.gains = {
            'px': gain_x,
            'py': gain_y,
        }

        self.slicer = xf.UniformBinSlicer(
            filling_scheme=filling_scheme,
            bunch_selection=bunch_selection,
            zeta_range=zeta_range,
            num_slices=num_slices,
            bunch_spacing_zeta=bunch_spacing_zeta,
            moments=['px', 'py']
        )

        if filling_scheme is not None:
            i_last_bunch = np.where(filling_scheme)[0][-1]
            num_periods = i_last_bunch + 1
        else:
            num_periods = 1

        self.moments_data = {}
        for moment in self.gains.keys():
            self.moments_data[moment] = CompressedProfile(
                moments=[moment],
                zeta_range=zeta_range,
                num_slices=num_slices,
                bunch_spacing_zeta=bunch_spacing_zeta,
                num_periods=num_periods,
                num_turns=1,
                circumference=circumference
            )

    def _reconfigure_for_parallel(self, n_procs, my_rank):
        filled_slots = self.slicer.filled_slots
        scheme = np.zeros(np.max(filled_slots) + 1,
                        dtype=np.int64)
        scheme[filled_slots] = 1

        split_scheme = xp.matched_gaussian.split_scheme
        bunch_selection_rank = split_scheme(filling_scheme=scheme,
                                             n_chunk=int(n_procs))

        self.slicer = xf.UniformBinSlicer(
            filling_scheme=scheme,
            bunch_selection=bunch_selection_rank[my_rank],
            zeta_range=self.slicer.zeta_range,
            num_slices=self.slicer.num_slices,
            bunch_spacing_zeta=self.slicer.bunch_spacing_zeta,
            moments=['px', 'py']
        )

    def track(self, particles, i_turn=0):
        i_slice_particles = particles.particle_id * 0 + -999
        i_slot_particles = particles.particle_id * 0 + -9999

        self.slicer.slice(particles, i_slice_particles=i_slice_particles,
                          i_slot_particles=i_slot_particles)

        for moment in ['px', 'py']:
            particles_moment = getattr(particles, moment)[:]
            slice_means = self.slicer.mean(moment)

            for i_bunch, bunch_number in enumerate(
                                self.slicer.bunch_selection):
                slot_mask = i_slot_particles == bunch_number
                slot_slices = np.unique(i_slice_particles[slot_mask])

                if len(self.slicer.bunch_selection) == 1:
                    slot_mean = np.mean(slice_means[slot_slices])
                else:
                    slot_mean = np.mean(slice_means[i_bunch, slot_slices])

                slot_mean = np.mean(particles_moment[slot_mask])

                particles_moment[slot_mask] -= (self.gains[moment] *
                                                slot_mean)



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from . import component
from . import devices
from . import element
from . import elements_group
from . import interface
from . import landau_damping
from . import materials
from . import model
from . import parameters
from . import plot
from . import sacherer_formula
from . import utilities
from . import utils

from .component import (Component, ComponentClassicThickWall, ComponentResonator,
                        ComponentTaperSingleLayerRestsistiveWall,
                        ComponentSingleLayerResistiveWall,
                        ComponentInterpolated,
                        ComponentFromArrays)
from .interface_dataclasses import Layer
from .element import Element


# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from typing import Tuple, List
import numpy as np

PI = np.pi

# Yokoya factors (taken from PyHEADTAIL)
_YOKOYA_FACTORS= {
    'circular': {'dipolar_x': 1.0, 'dipolar_y': 1.0,
                 'quadrupolar_x': 0.0,'quadrupolar_y': 0.0},
    'flat_horizontal': {'dipolar_x': PI**2 / 24, 'dipolar_y': PI**2 / 12,
                        'quadrupolar_x': -PI**2 / 24, 'quadrupolar_y': PI**2 / 24},
    'flat_vertical': {'dipolar_x': PI**2 / 12, 'dipolar_y': PI**2 / 24,
                      'quadrupolar_x': PI**2 / 24, 'quadrupolar_y': -PI**2 / 24}
}

class Yokoya:
    def __init__(self, shape: str | Tuple | List):
        self.shape = shape

        if isinstance(shape, str):
            assert shape in _YOKOYA_FACTORS
            self.dipolar_x = _YOKOYA_FACTORS[shape]['dipolar_x']
            self.dipolar_y = _YOKOYA_FACTORS[shape]['dipolar_y']
            self.quadrupolar_x = _YOKOYA_FACTORS[shape]['quadrupolar_x']
            self.quadrupolar_y = _YOKOYA_FACTORS[shape]['quadrupolar_y']
        else:
            assert len(shape) == 4
            self.dipolar_x = shape[0]
            self.dipolar_y = shape[1]
            self.quadrupolar_x = shape[2]
            self.quadrupolar_y = shape[3]

    def __getitem__(self, key):
        return getattr(self, key)

    def keys(self):
        return ['dipolar_x', 'dipolar_y', 'quadrupolar_x', 'quadrupolar_y']

    def __repr__(self):
        return f'Yokoya({self.shape})'



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from xtrack.pipeline.manager import PipelineManager
from xtrack.pipeline.multitracker import PipelineBranch, PipelineMultiTracker


def config_pipeline_for_wakes(particles, line, communicator,
                              elements_to_configure=None):

    assert communicator is not None, 'communicator must be provided'

    pipeline_manager=PipelineManager(communicator)

    comm_size = communicator.Get_size()
    my_rank = communicator.Get_rank()

    for rank in range(comm_size):
        pipeline_manager.add_particles(f'particles{rank}', rank)

    particles.init_pipeline(f'particles{my_rank}')

    if elements_to_configure is None:
        from xfields.beam_elements.element_with_slicer import ElementWithSlicer
        from xwakes.beam_elements.transverse_damper import TransverseDamper
        from xwakes.beam_elements.collective_monitor import CollectiveMonitor

        elements_to_configure = []
        for nn in line.element_names:
            if hasattr(line[nn], '_wake_tracker'):
                elements_to_configure.append(nn)
            elif isinstance(line[nn], ElementWithSlicer):
                elements_to_configure.append(nn)
            elif isinstance(line[nn], TransverseDamper):
                elements_to_configure.append(nn)
            elif isinstance(line[nn], CollectiveMonitor):
                elements_to_configure.append(nn)

    for nn in elements_to_configure:
        ee = line[nn]
        pipeline_manager.add_element(nn)
        if (isinstance(line[nn], TransverseDamper) or
            isinstance(line[nn], CollectiveMonitor)):
            line[nn]._reconfigure_for_parallel(comm_size, my_rank)
            continue

        if hasattr(ee, '_wake_tracker'):
            ee._reconfigure_for_parallel(comm_size, my_rank)
            ee = ee._wake_tracker

        ee.init_pipeline(
            pipeline_manager=pipeline_manager,
            element_name=nn,
            partner_names=[f'particles{rank}'
                            for rank in range(comm_size) if rank != my_rank])

    return pipeline_manager



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import pytest
import pathlib

import numpy as np
import pandas as pd

import xwakes as xw
import xobjects as xo

from scipy.constants import c as clight
from scipy.interpolate import interp1d

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

@pytest.mark.parametrize('wake_type', ['dipolar', 'quadrupolar'])
@pytest.mark.parametrize('plane', ['x', 'y'])
def test_impedances_vs_analytic_resonator_transverse(wake_type, plane):

    beta0 = 0.1

    wake = xw.WakeResonator(r=1e8, q=10, f_r=1e9,
        kind=f'{wake_type}_{plane}')

    assert len(wake.components) == 1
    assert wake.components[0].plane == plane
    assert wake.components[0].source_exponents == {
        'dipolar_x': (1, 0), 'dipolar_y': (0, 1),
        'quadrupolar_x': (0, 0), 'quadrupolar_y': (0, 0)}[f'{wake_type}_{plane}']
    assert wake.components[0].test_exponents == {
        'dipolar_x': (0, 0), 'dipolar_y': (0, 0),
        'quadrupolar_x': (1, 0), 'quadrupolar_y': (0, 1)}[f'{wake_type}_{plane}']

    z = np.linspace(-20, 20, 100000)
    t = np.linspace(-20/clight, 20/clight, 100000)

    w_vs_zeta = wake.components[0].function_vs_zeta(z, beta0=beta0, dzeta=1e-20)
    w_vs_t = wake.components[0].function_vs_t(t, beta0=beta0, dt=1e-20)

    # Assert that the function is positive at close to zero from the right
    assert wake.components[0].function_vs_t(1e-10, beta0=beta0, dt=1e-20) > 0
    assert wake.components[0].function_vs_t(-1e-10, beta0=beta0, dt=1e-20) == 0

    # Zeta has opposite sign compared to t
    assert wake.components[0].function_vs_zeta(-1e-3, beta0=beta0, dzeta=1e-20) > 0
    assert wake.components[0].function_vs_zeta(+1e-3, beta0=beta0, dzeta=1e-20) == 0

    omega = np.linspace(-10e9, 10e9, 500)

    Z_from_zeta = omega * (1 + 1j)
    Z_from_t = omega * (1 + 1j)
    dz = z[1] - z[0]
    dt = t[1] - t[0]
    for ii, oo in enumerate(omega):
        print(ii, end='\r', flush=True)
        Z_from_zeta[ii] = 1j/beta0/clight * np.sum(w_vs_zeta
                    * np.exp(1j * oo * z / beta0 / clight)) * dz
        Z_from_t[ii] = 1j * np.sum(
                    w_vs_t * np.exp(-1j * oo * t)) * dt

    Z_analytical = wake.components[0].impedance(omega/2/np.pi)

    f_test_sign = 1e9
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).real,
                    -wake.components[0].impedance(f_test_sign).real,
                    atol=0, rtol=1e-8)
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).imag,
                    wake.components[0].impedance(f_test_sign).imag,
                    atol=0, rtol=1e-8)

    xo.assert_allclose(
        Z_from_zeta, Z_analytical, rtol=0, atol=1e-4 * np.max(np.abs(Z_analytical)))
    xo.assert_allclose(
        Z_from_t, Z_analytical, rtol=0, atol=1e-4 * np.max(np.abs(Z_analytical)))


def test_impedances_vs_analytic_resonator_longitudinal():

    beta0 = 0.1

    wake = xw.WakeResonator(r=1e8, q=10, f_r=1e9,
        kind='longitudinal')

    assert len(wake.components) == 1
    assert wake.components[0].plane == 'z'
    assert wake.components[0].source_exponents == (0, 0)
    assert wake.components[0].test_exponents == (0, 0)

    # Assert that the function is positive at close to zero from the right
    assert wake.components[0].function_vs_t(1e-10, beta0=beta0, dt=1e-20) > 0
    assert wake.components[0].function_vs_t(-1e-10, beta0=beta0, dt=1e-20) == 0

    # Zeta has opposite sign compared to t
    assert wake.components[0].function_vs_zeta(-1e-3, beta0=beta0, dzeta=1e-20) > 0
    assert wake.components[0].function_vs_zeta(+1e-3, beta0=beta0, dzeta=1e-20) == 0

    z = np.linspace(-20, 20, 100000)
    t = np.linspace(-20/clight, 20/clight, 100000)

    w_vs_zeta = wake.components[0].function_vs_zeta(z, beta0=beta0, dzeta=1e-20)
    w_vs_t = wake.components[0].function_vs_t(t, beta0=beta0, dt=1e-20)

    omega = np.linspace(-10e9, 10e9, 500)

    Z_from_zeta = omega * (1 + 1j)
    Z_from_t = omega * (1 + 1j)
    dz = z[1] - z[0]
    dt = t[1] - t[0]
    for ii, oo in enumerate(omega):
        print(ii, end='\r', flush=True)
        Z_from_zeta[ii] = 1/beta0/clight * np.sum(w_vs_zeta
                    * np.exp(1j * oo * z / beta0 / clight)) * dz
        Z_from_t[ii] = np.sum(
                    w_vs_t * np.exp(-1j * oo * t)) * dt

    Z_analytical = wake.components[0].impedance(omega/2/np.pi)

    f_test_sign = 1e9
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).real,
                    wake.components[0].impedance(f_test_sign).real,
                    atol=0, rtol=1e-8)
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).imag,
                    -wake.components[0].impedance(f_test_sign).imag,
                    atol=0, rtol=1e-8)

    xo.assert_allclose(
        Z_from_zeta, Z_analytical, rtol=0, atol=1e-4 * np.max(np.abs(Z_analytical)))
    xo.assert_allclose(
        Z_from_t, Z_analytical, rtol=0, atol=1e-4 * np.max(np.abs(Z_analytical)))

    xo.assert_allclose(
        Z_from_zeta, Z_analytical, rtol=0, atol=1e-4 * np.max(np.abs(Z_analytical)))
    xo.assert_allclose(
        Z_from_t, Z_analytical, rtol=0, atol=1e-4 * np.max(np.abs(Z_analytical)))


@pytest.mark.parametrize('wake_type', ['dipolar'])
@pytest.mark.parametrize('plane', ['x', 'y'])
def test_impedances_vs_analytic_thick_res_wall_transverse(wake_type, plane):

    beta0 = 0.1

    wake = xw.WakeThickResistiveWall(
        kind=f'{wake_type}_{plane}',
        resistivity=1e-7,
        radius=1e-2,
        length=2.
        )

    assert len(wake.components) == 1
    assert wake.components[0].plane == plane
    assert wake.components[0].source_exponents == {
        'dipolar_x': (1, 0), 'dipolar_y': (0, 1),
        'quadrupolar_x': (0, 0), 'quadrupolar_y': (0, 0)}[f'{wake_type}_{plane}']
    assert wake.components[0].test_exponents == {
        'dipolar_x': (0, 0), 'dipolar_y': (0, 0),
        'quadrupolar_x': (1, 0), 'quadrupolar_y': (0, 1)}[f'{wake_type}_{plane}']

    z = np.linspace(-50, 50, 1000000)
    t = z / beta0 / clight

    w_vs_zeta = wake.components[0].function_vs_zeta(z, beta0=beta0, dzeta=1e-4)

    w_vs_t = wake.components[0].function_vs_t(t, beta0=beta0, dt=1e-4/beta0/clight)

    # Assert that the function is positive at close to zero from the right
    assert wake.components[0].function_vs_t(1e-10, beta0=beta0, dt=1e-20) > 0
    assert wake.components[0].function_vs_t(-1e-10, beta0=beta0, dt=1e-20) == 0

    # Zeta has opposite sign compared to t
    assert wake.components[0].function_vs_zeta(-1e-3, beta0=beta0, dzeta=1e-20) > 0
    assert wake.components[0].function_vs_zeta(+1e-3, beta0=beta0, dzeta=1e-20) == 0

    omega = np.linspace(-10e9, 10e9, 80)

    Z_from_zeta = omega * (1 + 1j)
    Z_from_t = omega * (1 + 1j)
    for ii, oo in enumerate(omega):
        print(ii, end='\r', flush=True)
        Z_from_zeta[ii] = 1j/beta0/clight * np.trapezoid(w_vs_zeta
                    * np.exp(1j * oo * z / beta0 / clight), z)
        Z_from_t[ii] = 1j * np.trapezoid(
                    w_vs_t * np.exp(-1j * oo * t), t)

    Z_analytical = wake.components[0].impedance(omega/2/np.pi)

    f_test_sign = 1e9
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).real,
                    -wake.components[0].impedance(f_test_sign).real,
                    atol=0, rtol=1e-8)
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).imag,
                    wake.components[0].impedance(f_test_sign).imag,
                    atol=0, rtol=1e-8)

    xo.assert_allclose(
        Z_from_zeta, Z_analytical, rtol=0, atol=0.1 * np.max(np.abs(Z_analytical)))
    xo.assert_allclose(
        Z_from_t, Z_analytical, rtol=0, atol=0.1 * np.max(np.abs(Z_analytical)))


    # Check that it scales with the length
    wake_len_default = xw.WakeThickResistiveWall(
        kind=f'{wake_type}_{plane}',
        resistivity=1e-7,
        radius=1e-2,
        )

    xo.assert_allclose(wake.components[0].wake(10.),
                       2 * wake_len_default.components[0].wake(10.),
                       rtol=0, atol=1e-10)


def test_impedances_vs_analytic_thick_res_wall_longitudinal():

    beta0 = 0.1

    wake = xw.WakeThickResistiveWall(
        kind='longitudinal',
        resistivity=1e-7,
        radius=1e-2,
        length=2.
        )

    assert len(wake.components) == 1
    assert wake.components[0].plane == 'z'
    assert wake.components[0].source_exponents == (0, 0)
    assert wake.components[0].test_exponents == (0, 0)

    # Assert that the function is positive at close to zero from the right
    assert wake.components[0].function_vs_t(1e-10, beta0=beta0, dt=1e-20) < 0 # unphysical, coming from asymptotic expansion
    assert wake.components[0].function_vs_t(-1e-10, beta0=beta0, dt=1e-20) == 0

    # Zeta has opposite sign compared to t
    assert wake.components[0].function_vs_zeta(-1e-3, beta0=beta0, dzeta=1e-20) < 0
    assert wake.components[0].function_vs_zeta(+1e-3, beta0=beta0, dzeta=1e-20) == 0

    z = np.linspace(-500, 500, 100000)
    t = np.linspace(-500/beta0/clight, 500/beta0/clight, 100000)

    w_vs_zeta = wake.components[0].function_vs_zeta(z, beta0=beta0, dzeta=1e-4)
    w_vs_t = wake.components[0].function_vs_t(t, beta0=beta0, dt=1e-4/beta0/clight)

    omega = np.linspace(-1e9, 1e9, 50)

    Z_from_zeta = omega * (1 + 1j)
    Z_from_t = omega * (1 + 1j)
    for ii, oo in enumerate(omega):
        print(ii, end='\r', flush=True)
        Z_from_zeta[ii] = 1/beta0/clight * np.trapezoid(w_vs_zeta
                    * np.exp(1j * oo * z / beta0 / clight), z)
        Z_from_t[ii] = np.trapezoid(
                    w_vs_t * np.exp(-1j * oo * t), t)

    Z_analytical = wake.components[0].impedance(omega/2/np.pi)

    Z_from_zeta -= (Z_from_zeta.mean() - Z_analytical.mean())
    Z_from_t -= (Z_from_t.mean() - Z_analytical.mean())

    f_test_sign = 1e9
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).real,
                    wake.components[0].impedance(f_test_sign).real,
                    atol=0, rtol=1e-8)
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).imag,
                    -wake.components[0].impedance(f_test_sign).imag,
                    atol=0, rtol=1e-8)

    xo.assert_allclose(
        Z_from_zeta, Z_analytical, rtol=0, atol=0.1 * np.max(np.abs(Z_analytical)))
    xo.assert_allclose(
        Z_from_t, Z_analytical, rtol=0, atol=0.1 * np.max(np.abs(Z_analytical)))

    xo.assert_allclose(
        Z_from_zeta, Z_analytical, rtol=0, atol=0.1 * np.max(np.abs(Z_analytical)))
    xo.assert_allclose(
        Z_from_t, Z_analytical, rtol=0, atol=0.1 * np.max(np.abs(Z_analytical)))

    # check that it scales with the length
    wake_len_default = xw.WakeThickResistiveWall(
        kind='longitudinal',
        resistivity=1e-7,
        radius=1e-2,
        )

    xo.assert_allclose(wake.components[0].wake(10.),
                          2 * wake_len_default.components[0].wake(10.),
                          rtol=0, atol=1e-10)

def test_impedances_vs_iw2d_vertical():

    beta0 = 0.1
    plane = 'y'
    wake_type = 'dipolar'

    wake_data = pd.read_csv(
        test_data_folder / 'iw2d_thick_wall/WydipWLHC_1layers10.00mm_precise.dat',
        skiprows=1, names=['z', 'wydip'], sep=' ')
    imp_data = pd.read_csv(
        test_data_folder / 'iw2d_thick_wall/ZydipWLHC_1layers10.00mm_precise.dat',
        skiprows=1, names=['f', 'ReZydip', 'ImZydip'], sep=' '
    )

    Z_first_quadrant = interp1d(imp_data['f'], imp_data['ReZydip'] + 1j * imp_data['ImZydip'])
    def Z_function(omega):
        isscalar = np.isscalar(omega)
        if isscalar:
            omega = np.array([omega])
        mask_zero = np.abs(omega) < imp_data['f'][0]
        if mask_zero.any():
            raise ValueError('Frequency too low')
        mask_positive = omega > 0
        mask_negative = omega < 0
        out = np.zeros_like(omega, dtype=complex)
        out[mask_positive] = Z_first_quadrant(omega[mask_positive])
        out[mask_negative] = Z_first_quadrant(-omega[mask_negative])
        out[mask_negative] = -np.conj(out[mask_negative])
        return out[0] if isscalar else out

    table = pd.DataFrame(
        {'time': wake_data['z'].values / clight, 'dipolar_y': wake_data['wydip'].values}
    )

    wake = xw.WakeFromTable(table=table)

    assert len(wake.components) == 1
    assert wake.components[0].plane == plane
    assert wake.components[0].source_exponents == {
        'dipolar_x': (1, 0), 'dipolar_y': (0, 1),
        'quadrupolar_x': (0, 0), 'quadrupolar_y': (0, 0)}[f'{wake_type}_{plane}']
    assert wake.components[0].test_exponents == {
        'dipolar_x': (0, 0), 'dipolar_y': (0, 0),
        'quadrupolar_x': (1, 0), 'quadrupolar_y': (0, 1)}[f'{wake_type}_{plane}']

    z = np.linspace(-50, 50, 1000000)
    t = z / beta0 / clight

    w_vs_zeta = wake.components[0].function_vs_zeta(z, beta0=beta0, dzeta=1e-20)
    w_vs_t = wake.components[0].function_vs_t(t, beta0=beta0, dt=1e-20)

    # Assert that the function is positive at close to zero from the right
    assert wake.components[0].function_vs_t(1e-10, beta0=beta0, dt=1e-20) > 0
    assert wake.components[0].function_vs_t(-1e-10, beta0=beta0, dt=1e-20) == 0

    # Zeta has opposite sign compared to t
    assert wake.components[0].function_vs_zeta(-1e-3, beta0=beta0, dzeta=1e-20) > 0
    assert wake.components[0].function_vs_zeta(+1e-3, beta0=beta0, dzeta=1e-20) == 0

    omega = np.linspace(-10e9, 10e9, 50)

    Z_from_zeta = omega * (1 + 1j)
    Z_from_t = omega * (1 + 1j)
    for ii, oo in enumerate(omega):
        print(ii, end='\r', flush=True)
        Z_from_zeta[ii] = 1j/beta0/clight * np.trapezoid(w_vs_zeta
                    * np.exp(1j * oo * z / beta0 / clight), z)
        Z_from_t[ii] = 1j * np.trapezoid(
                    w_vs_t * np.exp(-1j * oo * t), t)

    Z_analytical = Z_function(omega/2/np.pi)


    f_test_sign = 1e9
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).real,
                    -wake.components[0].impedance(f_test_sign).real,
                    atol=0, rtol=1e-8)
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).imag,
                    wake.components[0].impedance(f_test_sign).imag,
                    atol=0, rtol=1e-8)

    xo.assert_allclose(
        Z_from_zeta, Z_analytical, rtol=0, atol=0.05 * np.max(np.abs(Z_analytical)))
    xo.assert_allclose(
        Z_from_t, Z_analytical, rtol=0, atol=0.05 * np.max(np.abs(Z_analytical)))

def test_impedance_vs_iw2d_longitudinal():

    beta0 = 1.

    wake_data = pd.read_csv(
        test_data_folder / 'iw2d_thick_wall/WlongWLHC_1layers10.00mm_precise.dat',
        skiprows=1, names=['z', 'wlong'], sep=' ')
    imp_data = pd.read_csv(
        test_data_folder / 'iw2d_thick_wall/ZlongWLHC_1layers10.00mm_precise.dat',
        skiprows=1, names=['f', 'ReZlong', 'ImZlong'], sep=' '
    )

    Z_first_quadrant = interp1d(imp_data['f'], imp_data['ReZlong'] + 1j * imp_data['ImZlong'])
    def Z_function(omega):
        isscalar = np.isscalar(omega)
        if isscalar:
            omega = np.array([omega])
        mask_zero = np.abs(omega) < imp_data['f'][0]
        if mask_zero.any():
            raise ValueError('Frequency too low')
        mask_positive = omega >= 0
        mask_negative = omega < 0
        out = np.zeros_like(omega, dtype=complex)
        out[mask_positive] = Z_first_quadrant(omega[mask_positive])
        out[mask_negative] = Z_first_quadrant(-omega[mask_negative])
        out[mask_negative] = np.conj(out[mask_negative])
        return out[0] if isscalar else out

    table = pd.DataFrame(
        {'time': np.concatenate([[0], wake_data['z'].values / clight]),
        'longitudinal': np.concatenate([[wake_data['wlong'].values[0]], wake_data['wlong'].values])
        }
    )

    wake = xw.WakeFromTable(table=table)

    wake_thick = xw.WakeThickResistiveWall(
        kind='longitudinal',
        resistivity=1.7e-8, # Copper at room temperature
        radius=1e-2
        )

    assert len(wake.components) == 1
    assert wake.components[0].plane == 'z'
    assert wake.components[0].source_exponents == (0, 0)
    assert wake.components[0].test_exponents == (0, 0)

    # Assert that the function is positive at close to zero from the right
    assert wake.components[0].function_vs_t(1e-15, beta0=beta0, dt=1e-20) > 0
    assert wake.components[0].function_vs_t(-1e-15, beta0=beta0, dt=1e-20) == 0

    # Zeta has opposite sign compared to t
    assert wake.components[0].function_vs_zeta(-1e-5, beta0=beta0, dzeta=1e-20) > 0
    assert wake.components[0].function_vs_zeta(+1e-5, beta0=beta0, dzeta=1e-20) == 0

    z_positive = wake_data['z'].values
    z_negative = -wake_data['z'].values[::-1]
    z = np.concatenate((z_negative, z_positive))
    t = -z[::-1] / clight

    w_vs_zeta = wake.components[0].function_vs_zeta(z, beta0=beta0, dzeta=1e-200)
    w_vs_t = wake.components[0].function_vs_t(t, beta0=beta0, dt=1e-200/beta0/clight)
    w_thick_vs_t = wake_thick.components[0].function_vs_t(t, beta0=beta0, dt=1e-200/beta0/clight)

    omega = np.linspace(-10e9, 10e9, 500)

    Z_from_zeta = omega * (1 + 1j)
    Z_from_t = omega * (1 + 1j)
    Z_from_zeta_thick = omega * (1 + 1j)
    Z_from_t_thick = omega * (1 + 1j)
    dz = z[1] - z[0]
    dt = t[1] - t[0]
    for ii, oo in enumerate(omega):
        print(ii, end='\r', flush=True)
        Z_from_zeta[ii] = 1/beta0/clight * np.trapezoid(w_vs_zeta
                    * np.exp(1j * oo * (z / beta0 / clight)) , z)
        Z_from_t[ii] = np.trapezoid(
                    w_vs_t * np.exp(-1j * oo * t), t)
        Z_from_zeta_thick[ii] = 1/beta0/clight * np.trapezoid(w_vs_zeta
                    * np.exp(1j * oo * (z / beta0 / clight)) , z)
        Z_from_t_thick[ii] = np.trapezoid(
                    w_thick_vs_t * np.exp(-1j * oo * t), t)
    Z_analytical = Z_function(omega/2/np.pi)

    Z_thick_wall = wake_thick.components[0].impedance(omega/2/np.pi)

    Z_from_zeta -= (Z_from_zeta.mean() - Z_analytical.mean())
    Z_from_t -= (Z_from_t.mean() - Z_analytical.mean())
    Z_from_zeta_thick -= (Z_from_zeta_thick.mean() - Z_thick_wall.mean())
    Z_from_t_thick -= (Z_from_t_thick.mean() - Z_thick_wall.mean())

    f_test_sign = 1e9
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).real,
                    wake.components[0].impedance(f_test_sign).real,
                    atol=0, rtol=1e-8)
    xo.assert_allclose(wake.components[0].impedance(-f_test_sign).imag,
                    -wake.components[0].impedance(f_test_sign).imag,
                    atol=0, rtol=1e-8)

    xo.assert_allclose(
        Z_from_zeta, Z_analytical, rtol=0, atol=0.1 * np.max(np.abs(Z_analytical)))
    xo.assert_allclose(
        Z_from_t, Z_analytical, rtol=0, atol=0.1 * np.max(np.abs(Z_analytical)))

    mask_check_thick = (t > 1e-12) & (t < 2e-11)
    xo.assert_allclose(
        w_vs_t[mask_check_thick], w_thick_vs_t[mask_check_thick], rtol=0.05, atol=0)



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xtrack as xt
import xwakes as xw
from xobjects.test_helpers import for_all_test_contexts

class DummyCommunicator:
    def __init__(self, rank, size):
        self.rank = rank
        self.size = size

    def Get_size(self):
        return self.size

    def Get_rank(self):
        return self.rank

@for_all_test_contexts
def test_config_pipeline_for_wakes(test_context):
    comm_size = 3
    for my_rank in range(comm_size):
        communicator = DummyCommunicator(rank=my_rank, size=comm_size)

        zeta_range = (-1, 1)
        num_slices = 10
        filling_scheme = [1, 0, 1, 1, 1]
        bunch_spacing_zeta = 5
        num_turns = 1
        circumference = 100

        wake1 = xw.WakeResonator(kind='dipole_x', r=1e9, q=5, f_r=20e6)

        wake1.configure_for_tracking(
            zeta_range=zeta_range,
            num_slices=num_slices,
            filling_scheme=filling_scheme,
            bunch_spacing_zeta=bunch_spacing_zeta,
            num_turns=num_turns,
            circumference=circumference
        )

        wake2 = xw.WakeResonator(kind='dipole_x', r=1e9, q=5, f_r=20e6)

        wake2.configure_for_tracking(
            zeta_range=zeta_range,
            num_slices=num_slices,
            filling_scheme=filling_scheme,
            bunch_spacing_zeta=bunch_spacing_zeta,
            num_turns=num_turns,
            circumference=circumference
        )

        wake3 = wake1 + wake2

        wake3.configure_for_tracking(
            zeta_range=zeta_range,
            num_slices=num_slices,
            filling_scheme=filling_scheme,
            bunch_spacing_zeta=bunch_spacing_zeta,
            num_turns=num_turns,
            circumference=circumference
        )

        wake_no_config = xw.WakeResonator(kind='dipole_x', r=1e9, q=5, f_r=20e6)

        drift = xt.Drift(length=1)

        line = xt.Line(elements=[wake1, wake2, wake3, wake_no_config, drift])

        particles = xt.Particles(_context=test_context)

        pipeline_manager = xw.config_pipeline_for_wakes(particles, line, communicator,
                                    elements_to_configure=None)

        assert len(pipeline_manager._IDs) == comm_size
        assert len(pipeline_manager._particles_per_rank) == comm_size

        for ii in range(comm_size):
            assert list(pipeline_manager._IDs.keys())[ii] == f'particles{ii}'
            assert pipeline_manager._particles_per_rank[ii] == [f'particles{ii}']

        assert len(pipeline_manager._elements) == 3
        assert pipeline_manager._elements['e0'] == 0
        assert pipeline_manager._elements['e1'] == 1
        assert pipeline_manager._elements['e2'] == 2

        assert pipeline_manager._communicator == communicator
        assert pipeline_manager.verbose == False
        assert pipeline_manager._pending_requests == {}
        assert pipeline_manager._last_request_turn == {}

        assert wake1._wake_tracker.partner_names == [f'particles{rank}'
                                for rank in range(comm_size) if rank != my_rank]
        assert wake2._wake_tracker.partner_names == [f'particles{rank}'
                                for rank in range(comm_size) if rank != my_rank]
        assert wake3._wake_tracker.partner_names == [f'particles{rank}'
                                for rank in range(comm_size) if rank != my_rank]

        assert particles.name == f'particles{my_rank}'



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from typing import Tuple

from .basewake import BaseWake, _handle_kind
from .wit import ComponentClassicThickWall


class WakeThickResistiveWall(BaseWake):

    def __init__(self, kind: str = None,
                plane: str = None,
                source_exponents: Tuple[int, int] | None = None,
                test_exponents: Tuple[int, int] | None = None,
                radius: float = None,
                length: float = 1.0,
                resistivity: float = None):

        if kind is not None:

            kind = _handle_kind(kind)

            components = []
            for kk in kind.keys():
                ff = kind[kk]
                if ff == 0: # if the factor is zero we skip the component
                    continue
                cc = ComponentClassicThickWall(
                        radius=radius, resistivity=resistivity,
                        kind=kk,
                        factor=ff, plane=plane, length=length)
                components.append(cc)
        else:
            cc = ComponentClassicThickWall(
                        radius=radius, resistivity=resistivity,
                        plane=plane, source_exponents=source_exponents,
                        test_exponents=test_exponents,
                        factor=ff, length=length)
            components = [cc]

        self.components = components
        self.kind = kind



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from pathlib import Path

_pkg_root = Path(__file__).parent.absolute()




# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import numpy as np
from scipy.constants import c
from scipy.constants import physical_constants
from scipy.signal import hilbert
from scipy.stats import linregress

import xwakes as xw
import xtrack as xt
import xpart as xp
from xobjects.test_helpers import for_all_test_contexts


exclude_contexts = ['ContextPyopencl', 'ContextCupy']

@for_all_test_contexts(excluding=exclude_contexts)
def test_transverse_damper(test_context):
    longitudinal_mode = 'nonlinear'
    # Machine settings
    n_turns = 1000

    n_macroparticles = 10000
    intensity = 8e9

    alpha = 53.86**-2

    e0 = physical_constants['proton mass energy equivalent in MeV'][0]*1e6
    p0c = 450e9
    gamma = p0c/e0
    beta = np.sqrt(1-1/gamma**2)

    h_rf = 35640
    bunch_spacing_buckets = 10
    num_bunches = 2
    n_slices = 500

    epsn_x = 2e-6
    epsn_y = 2e-6
    taub = 0.9e-9
    sigma_z = taub*beta*c/4

    circumference = 26658.883

    momentum_compaction = alpha
    v_rf = 4e6
    f_rev = beta*c/circumference
    f_rf = f_rev*h_rf

    bucket_length = circumference / h_rf
    zeta_range = (-0.5*bucket_length, 0.5*bucket_length)

    filling_scheme = np.zeros(int(h_rf/bunch_spacing_buckets))
    filling_scheme[0: num_bunches] = 1
    filled_slots = np.nonzero(filling_scheme)[0]

    one_turn_map = xt.LineSegmentMap(
        length=circumference,
        betx=70, bety=80,
        qx=62.31, qy=60.32,
        longitudinal_mode=longitudinal_mode,
        voltage_rf=v_rf, frequency_rf=f_rf, lag_rf=180,
        momentum_compaction_factor=momentum_compaction,
    )

    gain_x = 0.01
    gain_y = 0.01

    transverse_damper = xw.TransverseDamper(
        gain_x=gain_x, gain_y=gain_y,
        filling_scheme=filling_scheme,
        zeta_range=zeta_range,
        num_slices=n_slices,
        bunch_spacing_zeta=bunch_spacing_buckets*bucket_length,
        circumference=circumference,
        mode='bunch-by-bunch'
    )

    line = xt.Line(elements=[[one_turn_map, transverse_damper][0]],
                   element_names=[['one_turn_map', 'transverse_damper'][0]])

    line.particle_ref = xt.Particles(p0c=p0c)
    line.build_tracker(_context=test_context)

    particles = xp.generate_matched_gaussian_multibunch_beam(
        _context=test_context,
        filling_scheme=filling_scheme,
        bunch_num_particles=n_macroparticles,
        bunch_intensity_particles=intensity,
        nemitt_x=epsn_x, nemitt_y=epsn_y, sigma_z=sigma_z,
        line=line, bunch_spacing_buckets=bunch_spacing_buckets,
        bunch_selection=filled_slots,
        rf_harmonic=[h_rf], rf_voltage=[v_rf],
        particle_ref=line.particle_ref,

    )

    # apply a distortion to the bunches
    amplitude = 1e-3
    particles.px += amplitude
    particles.py += amplitude

    mean_x = np.zeros((n_turns, num_bunches))
    mean_y = np.zeros((n_turns, num_bunches))

    for i_turn in range(n_turns):
        line.track(particles, num_turns=1)
        transverse_damper.track(particles, i_turn)
        for ib in range(num_bunches):
            mean_x[i_turn, ib] = np.mean(particles.x[n_macroparticles*ib:
                                                     n_macroparticles*(ib+1)])
            mean_y[i_turn, ib] = np.mean(particles.y[n_macroparticles*ib:
                                                     n_macroparticles*(ib+1)])

    turns = np.linspace(0, n_turns-1, n_turns)

    i_fit_start = 200
    i_fit_end = 600

    for i_bunch in range(num_bunches):
        ampls_x = np.abs(hilbert(mean_x[:, i_bunch]))
        fit_x = linregress(turns[i_fit_start: i_fit_end],
                           np.log(ampls_x[i_fit_start: i_fit_end]))

        assert np.isclose(-fit_x.slope*2, gain_x, atol=1e-4, rtol=0)

        ampls_y = np.abs(hilbert(mean_y[:, i_bunch]))
        fit_y = linregress(turns[i_fit_start: i_fit_end],
                           np.log(ampls_y[i_fit_start: i_fit_end]))

        assert np.isclose(-fit_y.slope*2, gain_y, atol=1e-4, rtol=0)



# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from dataclasses import dataclass
from typing import Tuple, Optional
import numpy as np

@dataclass(frozen=True, eq=True)
class Sampling:
    start: float
    stop: float
    # 0 = logarithmic, 1 = linear, 2 = both
    scan_type: int
    added: Tuple[float]
    sampling_exponent: Optional[float] = None
    points_per_decade: Optional[float] = None
    min_refine: Optional[float] = None
    max_refine: Optional[float] = None
    n_refine: Optional[float] = None


@dataclass(frozen=True, eq=True)
class Layer:
    # The distance in mm of the inner surface of the layer from the reference orbit
    thickness: float
    dc_resistivity: float
    resistivity_relaxation_time: float = 0.0
    re_dielectric_constant: float = 1.0
    magnetic_susceptibility: float = 0.0
    permeability_relaxation_frequency: float = np.inf


# Define several dataclasses for IW2D input elements. We must split mandatory
# and optional arguments into private dataclasses to respect the resolution
# order. The public classes RoundIW2DInput and FlatIW2D input inherit from
# from the private classes.
# https://stackoverflow.com/questions/51575931/class-inheritance-in-python-3-7-dataclasses

@dataclass(frozen=True, eq=True)
class _IW2DInputBase:
    machine: str
    length: float
    relativistic_gamma: float
    calculate_wake: bool
    f_params: Sampling


@dataclass(frozen=True, eq=True)
class _IW2DInputOptional:
    z_params: Optional[Sampling] = None
    long_factor: Optional[float] = None
    wake_tol: Optional[float] = None
    freq_lin_bisect: Optional[float] = None
    comment: Optional[str] = None


@dataclass(frozen=True, eq=True)
class IW2DInput(_IW2DInputOptional, _IW2DInputBase):
    pass


@dataclass(frozen=True, eq=True)
class _RoundIW2DInputBase(_IW2DInputBase):
    layers: Tuple[Layer]
    inner_layer_radius: float
    # (long, xdip, ydip, xquad, yquad)
    yokoya_factors: Tuple[float, float, float, float, float]


@dataclass(frozen=True, eq=True)
class _RoundIW2DInputOptional(_IW2DInputOptional):
    pass


@dataclass(frozen=True, eq=True)
class RoundIW2DInput(_RoundIW2DInputOptional, _RoundIW2DInputBase):
    pass


@dataclass(frozen=True, eq=True)
class _FlatIW2DInputBase(_IW2DInputBase):
    top_bottom_symmetry: bool
    top_layers: Tuple[Layer]
    top_half_gap: float


@dataclass(frozen=True, eq=True)
class _FlatIW2DInputOptional(_IW2DInputOptional):
    bottom_layers: Optional[Tuple[Layer]] = None
    bottom_half_gap: Optional[float] = None


@dataclass(frozen=True, eq=True)
class FlatIW2DInput(_FlatIW2DInputOptional, _FlatIW2DInputBase):
    pass